//Josip Medved <jmedved@jmedved.com> http://www.jmedved.com

//2009-03-15: Initial version.


using System;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.Collections.Generic;
using System.Globalization;

namespace Medo.IO {

    /// <summary>
    /// GPS Intermediate Driver communication.
    /// </summary>
    public class Gps : IDisposable {

        private IntPtr _handle;

        /// <summary>
        /// Creates new instance.
        /// </summary>
        public Gps() {
        }

        ~Gps() {
            Dispose(false);
        }


        /// <summary>
        /// Opens connection to GPS intermediate driver.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Device already open. -or- Cannot open device.</exception>
        public void Open() {
            if (this._handle != IntPtr.Zero) { throw new System.InvalidOperationException("Device already open."); }

            this._handle = NativeMethods.GPSOpenDevice(IntPtr.Zero, IntPtr.Zero, null, 0);
            if (this._handle == IntPtr.Zero) { throw new System.InvalidOperationException("Cannot open device."); }
        }

        /// <summary>
        /// Gets whether connection to GPS intermediate driver is open.
        /// </summary>
        public bool IsOpen {
            get { return (this._handle != IntPtr.Zero); }
        }

        /// <summary>
        /// Closes connection to GPS intermediate driver.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Cannot close device.</exception>
        public void Close() {
            if (this._handle != IntPtr.Zero) {
                if (NativeMethods.GPSCloseDevice(this._handle) == NativeMethods.ERROR_SUCCESS) {
                    this._handle = IntPtr.Zero;
                } else {
                    throw new System.InvalidOperationException("Cannot close device.");
                }
            }
        }


        /// <summary>
        /// Returns device state.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Cannot query GPS Intermediate driver.</exception>
        public static GpsDeviceState GetDeviceState() {
            NativeMethods.GPS_DEVICE pGPSDevice = new NativeMethods.GPS_DEVICE();
            pGPSDevice.dwVersion = 1;
            pGPSDevice.dwSize = Marshal.SizeOf(pGPSDevice);
            if (NativeMethods.GPSGetDeviceState(ref pGPSDevice) == NativeMethods.ERROR_SUCCESS) {
                return new GpsDeviceState((GpsState)pGPSDevice.dwServiceState, (GpsState)pGPSDevice.dwDeviceState, DateTime.FromFileTimeUtc(pGPSDevice.ftLastDataReceived), pGPSDevice.szGPSDriverPrefix, pGPSDevice.szGPSMultiplexPrefix, pGPSDevice.szGPSFriendlyName);
            } else {
                throw new System.InvalidOperationException("Cannot query GPS Intermediate driver.");
            }
        }


        /// <summary>
        /// Returns position data (maximum age 2500 ms). Valid only if GPS Intermediate Driver is open.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Connection to GPS Intermediate driver is not open. -or- Cannot query GPS Intermediate driver.</exception>
        public GpsPosition GetPosition() {
            return GetPosition(2500);
        }

        /// <summary>
        /// Returns position data. Valid only if GPS Intermediate Driver is open.
        /// </summary>
        /// <param name="maximumAge">Maximum age of data in milliseconds.</param>
        /// <exception cref="System.InvalidOperationException">Connection to GPS Intermediate driver is not open. -or- Cannot query GPS Intermediate driver.</exception>
        public GpsPosition GetPosition(int maximumAge) {
            if (!this.IsOpen) { throw new System.InvalidOperationException("Connection to GPS Intermediate driver is not open."); }

            var pGPSPosition = new NativeMethods.GPS_POSITION();
            pGPSPosition.dwVersion = 1;
            pGPSPosition.dwSize = Marshal.SizeOf(pGPSPosition);
            if (NativeMethods.GPSGetPosition(this._handle, ref pGPSPosition, maximumAge, 0) == NativeMethods.ERROR_SUCCESS) {
                DateTime? utcTime = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_UTC_TIME) == NativeMethods.GPS_VALID_UTC_TIME) { utcTime = new DateTime(pGPSPosition.stUTCTime.wYear, pGPSPosition.stUTCTime.wMonth, pGPSPosition.stUTCTime.wDay, pGPSPosition.stUTCTime.wHour, pGPSPosition.stUTCTime.wMinute, pGPSPosition.stUTCTime.wSecond, pGPSPosition.stUTCTime.wMilliseconds, DateTimeKind.Utc); }
                double? latitude = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_LATITUDE) == NativeMethods.GPS_VALID_LATITUDE) { latitude = pGPSPosition.dblLatitude; }
                double? longitude = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_LONGITUDE) == NativeMethods.GPS_VALID_LONGITUDE) { longitude = pGPSPosition.dblLongitude; }
                float? speed = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_SPEED) == NativeMethods.GPS_VALID_SPEED) { speed = pGPSPosition.flSpeed; }
                float? heading = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_HEADING) == NativeMethods.GPS_VALID_HEADING) { heading = pGPSPosition.flHeading; }
                double? magneticVariation = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_MAGNETIC_VARIATION) == NativeMethods.GPS_VALID_MAGNETIC_VARIATION) { magneticVariation = pGPSPosition.dblMagneticVariation; }
                float? altitudeWRTSeaLevel = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_ALTITUDE_WRT_SEA_LEVEL) == NativeMethods.GPS_VALID_ALTITUDE_WRT_SEA_LEVEL) { altitudeWRTSeaLevel = pGPSPosition.flAltitudeWRTSeaLevel; }
                float? altitudeWRTEllipsoid = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_ALTITUDE_WRT_ELLIPSOID) == NativeMethods.GPS_VALID_ALTITUDE_WRT_ELLIPSOID) { altitudeWRTEllipsoid = pGPSPosition.flAltitudeWRTEllipsoid; }
                float? positionDilutionOfPrecision = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_POSITION_DILUTION_OF_PRECISION) == NativeMethods.GPS_VALID_POSITION_DILUTION_OF_PRECISION) { positionDilutionOfPrecision = pGPSPosition.flPositionDilutionOfPrecision; }
                float? horizontalDilutionOfPrecision = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_HORIZONTAL_DILUTION_OF_PRECISION) == NativeMethods.GPS_VALID_HORIZONTAL_DILUTION_OF_PRECISION) { horizontalDilutionOfPrecision = pGPSPosition.flHorizontalDilutionOfPrecision; }
                float? verticalDilutionOfPrecision = null; if ((pGPSPosition.dwValidFields & NativeMethods.GPS_VALID_VERTICAL_DILUTION_OF_PRECISION) == NativeMethods.GPS_VALID_VERTICAL_DILUTION_OF_PRECISION) { verticalDilutionOfPrecision = pGPSPosition.flVerticalDilutionOfPrecision; }

                var satellites = new Dictionary<int, GpsSatellite>();
                if (pGPSPosition.rgdwSatellitesInViewPRNs01 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs01, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs01, pGPSPosition.rgdwSatellitesInViewElevation01, pGPSPosition.rgdwSatellitesInViewAzimuth01, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio01, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs01, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs02 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs02, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs02, pGPSPosition.rgdwSatellitesInViewElevation02, pGPSPosition.rgdwSatellitesInViewAzimuth02, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio02, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs02, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs03 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs03, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs03, pGPSPosition.rgdwSatellitesInViewElevation03, pGPSPosition.rgdwSatellitesInViewAzimuth03, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio03, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs03, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs04 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs04, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs04, pGPSPosition.rgdwSatellitesInViewElevation04, pGPSPosition.rgdwSatellitesInViewAzimuth04, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio04, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs04, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs05 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs05, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs05, pGPSPosition.rgdwSatellitesInViewElevation05, pGPSPosition.rgdwSatellitesInViewAzimuth05, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio05, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs05, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs06 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs06, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs06, pGPSPosition.rgdwSatellitesInViewElevation06, pGPSPosition.rgdwSatellitesInViewAzimuth06, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio06, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs06, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs07 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs07, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs07, pGPSPosition.rgdwSatellitesInViewElevation07, pGPSPosition.rgdwSatellitesInViewAzimuth07, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio07, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs07, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs08 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs08, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs08, pGPSPosition.rgdwSatellitesInViewElevation08, pGPSPosition.rgdwSatellitesInViewAzimuth08, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio08, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs08, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs09 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs09, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs09, pGPSPosition.rgdwSatellitesInViewElevation09, pGPSPosition.rgdwSatellitesInViewAzimuth09, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio09, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs09, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs10 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs10, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs10, pGPSPosition.rgdwSatellitesInViewElevation10, pGPSPosition.rgdwSatellitesInViewAzimuth10, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio10, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs10, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs11 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs11, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs11, pGPSPosition.rgdwSatellitesInViewElevation11, pGPSPosition.rgdwSatellitesInViewAzimuth11, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio11, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs11, pGPSPosition))); }
                if (pGPSPosition.rgdwSatellitesInViewPRNs12 != 0) { satellites.Add(pGPSPosition.rgdwSatellitesInViewPRNs12, new GpsSatellite(pGPSPosition.rgdwSatellitesInViewPRNs12, pGPSPosition.rgdwSatellitesInViewElevation12, pGPSPosition.rgdwSatellitesInViewAzimuth12, pGPSPosition.rgdwSatellitesInViewSignalToNoiseRatio12, IsSatelliteUsed(pGPSPosition.rgdwSatellitesInViewPRNs12, pGPSPosition))); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs01 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs01)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs01, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs01, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs02 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs02)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs02, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs02, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs03 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs03)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs03, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs03, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs04 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs04)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs04, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs04, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs05 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs05)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs05, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs05, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs06 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs06)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs06, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs06, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs07 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs07)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs07, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs07, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs08 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs08)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs08, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs08, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs09 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs09)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs09, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs09, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs10 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs10)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs10, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs10, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs11 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs11)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs11, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs11, 0, 0, -1, true)); }
                if ((pGPSPosition.rgdwSatellitesUsedPRNs12 != 0) && !satellites.ContainsKey(pGPSPosition.rgdwSatellitesUsedPRNs12)) { satellites.Add(pGPSPosition.rgdwSatellitesUsedPRNs12, new GpsSatellite(pGPSPosition.rgdwSatellitesUsedPRNs12, 0, 0, -1, true)); }

                GpsSatellite[] satellitesArray = new GpsSatellite[satellites.Count];
                int iSatellitesIndex = 0;
                foreach (var iItem in satellites) {
                    satellitesArray[iSatellitesIndex] = iItem.Value;
                    iSatellitesIndex += 1;
                }

                return new GpsPosition(utcTime, latitude, longitude, speed, heading, magneticVariation, altitudeWRTSeaLevel, altitudeWRTEllipsoid, (GpsFixQuality)pGPSPosition.FixQuality, (GpsFixType)pGPSPosition.FixType, (GpsFixSelectionType)pGPSPosition.SelectionType, positionDilutionOfPrecision, horizontalDilutionOfPrecision, verticalDilutionOfPrecision, satellitesArray);
            } else {
                throw new System.InvalidOperationException("Cannot query GPS Intermediate driver.");
            }
        }


        private static bool IsSatelliteUsed(int prn, NativeMethods.GPS_POSITION positionData) {
            if (prn == positionData.rgdwSatellitesUsedPRNs01) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs02) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs03) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs04) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs05) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs06) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs07) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs08) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs09) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs10) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs11) { return true; }
            if (prn == positionData.rgdwSatellitesUsedPRNs12) { return true; }
            return false;
        }


        #region IDisposable Members

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        /// <param name="disposing">True if managed resources should be disposed; otherwise, false.</param>
        protected virtual void Dispose(bool disposing) {
            if (disposing) { }
            this.Close();
        }


        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        public void Dispose() {
            Dispose(true);
            System.GC.SuppressFinalize(this);
        }

        #endregion


        private static class NativeMethods {

            internal const Int32 ERROR_SUCCESS = 0;
            internal const Int32 GPS_VALID_UTC_TIME = 0x00000001;
            internal const Int32 GPS_VALID_LATITUDE = 0x00000002;
            internal const Int32 GPS_VALID_LONGITUDE = 0x00000004;
            internal const Int32 GPS_VALID_SPEED = 0x00000008;
            internal const Int32 GPS_VALID_HEADING = 0x00000010;
            internal const Int32 GPS_VALID_MAGNETIC_VARIATION = 0x00000020;
            internal const Int32 GPS_VALID_ALTITUDE_WRT_SEA_LEVEL = 0x00000040;
            internal const Int32 GPS_VALID_ALTITUDE_WRT_ELLIPSOID = 0x00000080;
            internal const Int32 GPS_VALID_POSITION_DILUTION_OF_PRECISION = 0x00000100;
            internal const Int32 GPS_VALID_HORIZONTAL_DILUTION_OF_PRECISION = 0x00000200;
            internal const Int32 GPS_VALID_VERTICAL_DILUTION_OF_PRECISION = 0x00000400;
            internal const Int32 GPS_VALID_SATELLITE_COUNT = 0x00000800;
            internal const Int32 GPS_VALID_SATELLITES_USED_PRNS = 0x00001000;
            internal const Int32 GPS_VALID_SATELLITES_IN_VIEW = 0x00002000;
            internal const Int32 GPS_VALID_SATELLITES_IN_VIEW_PRNS = 0x00004000;
            internal const Int32 GPS_VALID_SATELLITES_IN_VIEW_ELEVATION = 0x00008000;
            internal const Int32 GPS_VALID_SATELLITES_IN_VIEW_AZIMUTH = 0x000010000;
            internal const Int32 GPS_VALID_SATELLITES_IN_VIEW_SIGNAL_TO_NOISE_RATIO = 0x00020000;


            [StructLayout(LayoutKind.Sequential)]
            internal struct FILETIME {
                internal Int32 dwLowDateTime;
                internal Int32 dwHighDateTime;
            }

            [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
            internal struct GPS_DEVICE {
                internal Int32 dwVersion;
                internal Int32 dwSize;
                internal Int32 dwServiceState;
                internal Int32 dwDeviceState;
                internal Int64 ftLastDataReceived;
                [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 16)]
                internal string szGPSDriverPrefix;
                [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 16)]
                internal string szGPSMultiplexPrefix;
                [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 64)]
                internal string szGPSFriendlyName;
            }

            [StructLayout(LayoutKind.Sequential)]
            internal struct SYSTEMTIME {
                internal Int16 wYear;
                internal Int16 wMonth;
                internal Int16 wDayOfWeek;
                internal Int16 wDay;
                internal Int16 wHour;
                internal Int16 wMinute;
                internal Int16 wSecond;
                internal Int16 wMilliseconds;
            }


            [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
            internal struct GPS_POSITION {
                internal Int32 dwVersion;
                internal Int32 dwSize;
                internal Int32 dwValidFields;
                internal Int32 dwFlags;
                internal SYSTEMTIME stUTCTime;
                internal Double dblLatitude;
                internal Double dblLongitude;
                internal Single flSpeed;
                internal Single flHeading;
                internal Double dblMagneticVariation;
                internal Single flAltitudeWRTSeaLevel;
                internal Single flAltitudeWRTEllipsoid;
                internal Int32 FixQuality;
                internal Int32 FixType;
                internal Int32 SelectionType;
                internal Single flPositionDilutionOfPrecision;
                internal Single flHorizontalDilutionOfPrecision;
                internal Single flVerticalDilutionOfPrecision;

                internal Int32 dwSatelliteCount;
                internal Int32 rgdwSatellitesUsedPRNs01;
                internal Int32 rgdwSatellitesUsedPRNs02;
                internal Int32 rgdwSatellitesUsedPRNs03;
                internal Int32 rgdwSatellitesUsedPRNs04;
                internal Int32 rgdwSatellitesUsedPRNs05;
                internal Int32 rgdwSatellitesUsedPRNs06;
                internal Int32 rgdwSatellitesUsedPRNs07;
                internal Int32 rgdwSatellitesUsedPRNs08;
                internal Int32 rgdwSatellitesUsedPRNs09;
                internal Int32 rgdwSatellitesUsedPRNs10;
                internal Int32 rgdwSatellitesUsedPRNs11;
                internal Int32 rgdwSatellitesUsedPRNs12;

                internal Int32 dwSatellitesInView;
                internal Int32 rgdwSatellitesInViewPRNs01;
                internal Int32 rgdwSatellitesInViewPRNs02;
                internal Int32 rgdwSatellitesInViewPRNs03;
                internal Int32 rgdwSatellitesInViewPRNs04;
                internal Int32 rgdwSatellitesInViewPRNs05;
                internal Int32 rgdwSatellitesInViewPRNs06;
                internal Int32 rgdwSatellitesInViewPRNs07;
                internal Int32 rgdwSatellitesInViewPRNs08;
                internal Int32 rgdwSatellitesInViewPRNs09;
                internal Int32 rgdwSatellitesInViewPRNs10;
                internal Int32 rgdwSatellitesInViewPRNs11;
                internal Int32 rgdwSatellitesInViewPRNs12;
                internal Int32 rgdwSatellitesInViewElevation01;
                internal Int32 rgdwSatellitesInViewElevation02;
                internal Int32 rgdwSatellitesInViewElevation03;
                internal Int32 rgdwSatellitesInViewElevation04;
                internal Int32 rgdwSatellitesInViewElevation05;
                internal Int32 rgdwSatellitesInViewElevation06;
                internal Int32 rgdwSatellitesInViewElevation07;
                internal Int32 rgdwSatellitesInViewElevation08;
                internal Int32 rgdwSatellitesInViewElevation09;
                internal Int32 rgdwSatellitesInViewElevation10;
                internal Int32 rgdwSatellitesInViewElevation11;
                internal Int32 rgdwSatellitesInViewElevation12;
                internal Int32 rgdwSatellitesInViewAzimuth01;
                internal Int32 rgdwSatellitesInViewAzimuth02;
                internal Int32 rgdwSatellitesInViewAzimuth03;
                internal Int32 rgdwSatellitesInViewAzimuth04;
                internal Int32 rgdwSatellitesInViewAzimuth05;
                internal Int32 rgdwSatellitesInViewAzimuth06;
                internal Int32 rgdwSatellitesInViewAzimuth07;
                internal Int32 rgdwSatellitesInViewAzimuth08;
                internal Int32 rgdwSatellitesInViewAzimuth09;
                internal Int32 rgdwSatellitesInViewAzimuth10;
                internal Int32 rgdwSatellitesInViewAzimuth11;
                internal Int32 rgdwSatellitesInViewAzimuth12;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio01;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio02;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio03;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio04;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio05;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio06;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio07;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio08;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio09;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio10;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio11;
                internal Int32 rgdwSatellitesInViewSignalToNoiseRatio12;
            }

            [DllImport("gpsapi.dll")]
            internal static extern IntPtr GPSOpenDevice(IntPtr hNewLocationData, IntPtr hDeviceStateChange, String szDeviceName, Int32 dwFlags);

            [DllImport("gpsapi.dll")]
            internal static extern Int32 GPSCloseDevice(IntPtr hGPSDevice);

            [DllImport("gpsapi.dll")]
            internal static extern Int32 GPSGetDeviceState(ref GPS_DEVICE pGPSDevice);

            [DllImport("gpsapi.dll")]
            internal static extern Int32 GPSGetPosition(IntPtr hGPSDevice, ref GPS_POSITION pGPSPosition, Int32 dwMaximumAge, Int32 dwFlags);

        }

    }



    /// <summary>
    /// State of GPS Intermediate driver.
    /// </summary>
    public class GpsDeviceState {

        /// <summary>
        /// Creates new instance.
        /// </summary>
        /// <param name="serviceState">State of the GPS Intermediate Driver.</param>
        /// <param name="deviceState">State of the driver controlling the GPS hardware.</param>
        /// <param name="lastDataReceived">Most recent time the GPS Intermediate Driver received information from the GPS hardware.</param>
        /// <param name="driverPrefix">Prefix string used to open the GPS hardware.</param>
        /// <param name="multiplexPrefix">Prefix string used to open the GPS Intermediate Driver multiplexer.</param>
        /// <param name="friendlyName">Friendly name of the GPS hardware currently used by the GPS Intermediate Driver.</param>
        internal GpsDeviceState(GpsState serviceState, GpsState deviceState, DateTime lastDataReceived, string driverPrefix, string multiplexPrefix, string friendlyName) {
            this.ServiceState = serviceState;
            this.DeviceState = deviceState;
            this.LastDataReceived = lastDataReceived;
            this.DriverPrefix = driverPrefix;
            this.MultiplexPrefix = multiplexPrefix;
            this.FriendlyName = friendlyName;
        }

        /// <summary>
        /// Gets state of the GPS Intermediate Driver.
        /// </summary>
        public GpsState ServiceState { get; private set; }

        /// <summary>
        /// Gets state of the driver controlling the GPS hardware
        /// </summary>
        public GpsState DeviceState { get; private set; }

        /// <summary>
        /// Gets most recent time the GPS Intermediate Driver received information from the GPS hardware. This time is based on UTC according to the local system clock, not the clock information used by GPS satellites.
        /// </summary>
        public DateTime LastDataReceived { get; private set; }

        /// <summary>
        /// Gets prefix string used to open the GPS hardware. This entry is empty when the GPS Intermediate Driver reads information from static files.
        /// </summary>
        public string DriverPrefix { get; private set; }

        /// <summary>
        /// Gets prefix string used to open the GPS Intermediate Driver multiplexer.
        /// </summary>
        public string MultiplexPrefix { get; private set; }

        /// <summary>
        /// Gets friendly name of the GPS hardware currently used by the GPS Intermediate Driver.
        /// </summary>
        public string FriendlyName { get; private set; }

    }



    /// <summary>
    /// States of GPS service and device.
    /// </summary>
    public enum GpsState : int {
        /// <summary>
        /// The service is turned off.
        /// </summary>
        Off = 0,
        /// <summary>
        /// The service is turned on.
        /// </summary>
        On = 1,
        /// <summary>
        /// The service is in the process of starting up.
        /// </summary>
        StartingUp = 2,
        /// <summary>
        /// The service is in the process of shutting down.
        /// </summary>
        ShuttingDown = 3,
        /// <summary>
        /// The service is in the process of unloading.
        /// </summary>
        Unloading = 4,
        /// <summary>
        /// The service is not initialized.
        /// </summary>
        Uninitialized = 5,
        /// <summary>
        /// Status is unknown
        /// </summary>
        Unknown = -1,
    }



    /// <summary>
    /// GPS position details.
    /// </summary>
    public class GpsPosition {

        private GpsSatellite[] _satellites;

        /// <summary>
        /// Creates new instance.
        /// </summary>
        /// <param name="utcTime">Universal time (UTC) according to information provided by GPS satellites.</param>
        /// <param name="latitude">Latitude, in degrees. Positive numbers indicate north latitude.</param>
        /// <param name="longitude">Longitude, in degrees. Positive numbers indicate east longitude.</param>
        /// <param name="speed">Speed, in knots (nautical miles).</param>
        /// <param name="heading">Heading, in degrees. A heading of zero is true north.</param>
        /// <param name="magneticVariation">Magnetic variation, which is the difference between the bearing to true north and the bearing shown on a magnetic compass. Positive numbers indicate east.</param>
        /// <param name="altitudeWRTSeaLevel">Altitude, in meters, with respect to sea level. </param>
        /// <param name="altitudeWRTEllipsoid">Altitude, in meters, with respect to the WGS84 ellipsoid.</param>
        /// <param name="fixQuality">Quality of the GPS fix.</param>
        /// <param name="fixType">Type of GPS fix.</param>
        /// <param name="fixSelectionType">Whether 2-D or 3-D mode is selected automatically or manually.</param>
        /// <param name="positionDilutionOfPrecision">Degree to which the overall position is affected by positional dilution of position (PDOP). PDOP is caused by the location of the satellites providing the GPS fix. Lower numbers indicate a more accurate position. A value of 1.0 indicates the least dilution (highest accuracy); a value of 50.0 indicates the most dilution (lowest accuracy).</param>
        /// <param name="horizontalDilutionOfPrecision">Degree to which the horizontal position (latitude and longitude) is affected by horizontal dilution of position (HDOP). HDOP is caused by the location of the satellites providing the GPS fix. Lower numbers indicate a more accurate position. A value of 1.0 indicates the least dilution (highest accuracy); a value of 50.0 indicates the most dilution (lowest accuracy).</param>
        /// <param name="verticalDilutionOfPrecision">Degree to which the vertical position (altitude) is affected by vertical dilution of position (VDOP). VDOP is caused by the location of the satellites providing the GPS fix. Lower numbers indicate a more accurate position. A value of 1.0 indicates the least dilution (highest accuracy); a value of 50.0 indicates the most dilution (lowest accuracy).</param>
        /// <param name="satellites">Satellites in view.</param>
        internal GpsPosition(DateTime? utcTime, double? latitude, double? longitude, float? speed, float? heading, double? magneticVariation, float? altitudeWRTSeaLevel, float? altitudeWRTEllipsoid, GpsFixQuality fixQuality, GpsFixType fixType, GpsFixSelectionType fixSelectionType, float? positionDilutionOfPrecision, float? horizontalDilutionOfPrecision, float? verticalDilutionOfPrecision, GpsSatellite[] satellites) {
            this.UtcTime = utcTime;
            this.Latitude = latitude;
            this.Longitude = longitude;
            this._speedInKnots = speed;
            this.Heading = heading;
            this.MagneticVariation = magneticVariation;
            this.Altitude = altitudeWRTSeaLevel;
            this.Wgs84EllipsoidAltitude = altitudeWRTEllipsoid;
            this.FixQuality = fixQuality;
            this.FixType = fixType;
            this.FixSelectionType = fixSelectionType;
            this.PositionDilutionOfPrecision = positionDilutionOfPrecision;
            this.HorizontalDilutionOfPrecision = horizontalDilutionOfPrecision;
            this.VerticalDilutionOfPrecision = verticalDilutionOfPrecision;
            this._satellites = satellites;
        }

        /// <summary>
        /// Gets universal time (UTC) according to information provided by GPS satellites.
        /// </summary>
        public DateTime? UtcTime { get; private set; }

        /// <summary>
        /// Gets latitude in degrees. Positive numbers indicate north latitude.
        /// </summary>
        public double? Latitude { get; private set; }

        /// <summary>
        /// Gets longitude in degrees. Positive numbers indicate east longitude.
        /// </summary>
        public double? Longitude { get; private set; }

        private float? _speedInKnots;
        /// <summary>
        /// Gets speed in meters per second.
        /// </summary>
        public float? Speed {
            get { return this._speedInKnots * 0.514F; }
        }

        /// <summary>
        /// Gets heading, in degrees. A heading of zero is true north.
        /// </summary>
        public float? Heading { get; private set; }

        /// <summary>
        /// Gets magnetic variation in degrees. Positive numbers indicate east.
        /// </summary>
        public double? MagneticVariation { get; private set; }

        /// <summary>
        /// Altitude in meters, with respect to sea level.
        /// </summary>
        public float? Altitude { get; private set; }

        /// <summary>
        /// Gets altitude in meters, with respect to the WGS84 ellipsoid.
        /// </summary>
        public float? Wgs84EllipsoidAltitude { get; private set; }

        /// <summary>
        /// Gets quality of the GPS fix.
        /// </summary>
        public GpsFixQuality FixQuality { get; private set; }

        /// <summary>
        /// Gets type of GPS fix.
        /// </summary>
        public GpsFixType FixType { get; private set; }

        /// <summary>
        /// Gets whether 2-D or 3-D mode is selected automatically or manually.
        /// </summary>
        public GpsFixSelectionType FixSelectionType { get; private set; }

        /// <summary>
        /// Gets degree to which the overall position is affected by positional dilution of position (PDOP). PDOP is caused by the location of the satellites providing the GPS fix. Lower numbers indicate a more accurate position. A value of 1.0 indicates the least dilution (highest accuracy); a value of 50.0 indicates the most dilution (lowest accuracy).
        /// </summary>
        public float? PositionDilutionOfPrecision { get; private set; }

        /// <summary>
        /// Gets degree to which the horizontal position (latitude and longitude) is affected by horizontal dilution of position (HDOP). HDOP is caused by the location of the satellites providing the GPS fix. Lower numbers indicate a more accurate position. A value of 1.0 indicates the least dilution (highest accuracy); a value of 50.0 indicates the most dilution (lowest accuracy). 
        /// </summary>
        public float? HorizontalDilutionOfPrecision { get; private set; }

        /// <summary>
        /// Gets degree to which the vertical position (altitude) is affected by vertical dilution of position (VDOP). VDOP is caused by the location of the satellites providing the GPS fix. Lower numbers indicate a more accurate position. A value of 1.0 indicates the least dilution (highest accuracy); a value of 50.0 indicates the most dilution (lowest accuracy).
        /// </summary>
        public float? VerticalDilutionOfPrecision { get; private set; }


        /// <summary>
        /// Gets number of satellites used for obtaining positon.
        /// </summary>
        public int SatellitesCount {
            get {
                if (this._satellites == null) { return 0; }
                int count = 0;
                foreach (var iSatellite in this._satellites) {
                    if (iSatellite.IsUsed) { count += 1; }
                }
                return count;
            }
        }

        /// <summary>
        /// Returns total number of satellites in view.
        /// </summary>
        public int SatellitesInViewCount {
            get {
                if (this._satellites == null) { return 0; }
                return this._satellites.Length;
            }
        }

        /// <summary>
        /// Returns all satellites in view. Returns null if no data is available.
        /// </summary>
        public GpsSatellite[] GetSatellites() {
            return this._satellites;
        }

    }


    /// <summary>
    /// Satellite details.
    /// </summary>
    public class GpsSatellite : IComparable<GpsSatellite> {

        /// <summary>
        /// Creates new instance.
        /// </summary>
        /// <param name="prn">PRN number of the satellite.</param>
        /// <param name="elevation">Elevation, in degrees, of the satellite.</param>
        /// <param name="azimuth">Azimuth, in degrees, of the satellite.</param>
        /// <param name="signalToNoiseRatio">Signal to noise ratio of the satellite.</param>
        /// <param name="isUsed">True if satellite is used for obtaining position data.</param>
        internal GpsSatellite(int prn, int elevation, int azimuth, int signalToNoiseRatio, bool isUsed) {
            this.Prn = prn;
            this.Elevation = elevation;
            this.Azimuth = azimuth;
            this.SignalToNoiseRatio = signalToNoiseRatio;
            this.IsUsed = isUsed;
        }

        /// <summary>
        /// PRN (pseudo-random noise) number of the satellite.
        /// </summary>
        public int Prn { get; private set; }

        /// <summary>
        /// Elevation, in degrees, of the satellite.
        /// </summary>
        public int Elevation { get; private set; }

        /// <summary>
        /// Azimuth, in degrees, of the satellite.
        /// </summary>
        public int Azimuth { get; private set; }

        /// <summary>
        /// Gets signal to noise ratio of the satellite. Higher numbers indicate greater signal strength.
        /// </summary>
        public int SignalToNoiseRatio { get; private set; }

        /// <summary>
        /// Gets whether satellite is used to obtain position data.
        /// </summary>
        public bool IsUsed { get; private set; }


        /// <summary>
        /// Determines whether the specified System.Object is equal to the current System.Object.
        /// </summary>
        /// <param name="obj">The System.Object to compare with the current System.Object.</param>
        public override bool Equals(object obj) {
            GpsSatellite other = obj as GpsSatellite;
            if ((other != null) && (this.Prn == other.Prn)) { return true; }

            return false;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        public override int GetHashCode() {
            return this.Prn.GetHashCode();
        }

        /// <summary>
        /// Returns a System.String that represents the current System.Object.
        /// </summary>
        public override string ToString() {
            return this.Prn.ToString(CultureInfo.InvariantCulture);
        }


        #region IComparable<GpsSatellite> Members

        /// <summary>
        /// Compares the current object with another object of the same type.
        /// </summary>
        /// <param name="other">An object to compare with this object.</param>
        int IComparable<GpsSatellite>.CompareTo(GpsSatellite other) {
            return this.Prn - other.Prn;
        }

        #endregion
    }


    /// <summary>
    /// Quality of the GPS fix.
    /// </summary>
    public enum GpsFixQuality : int {

        /// <summary>
        /// Quality of fix is unknown. 
        /// </summary>
        Unknown = 0,

        /// <summary>
        /// Fix uses information from GPS satellites only. 
        /// </summary>
        Gps,

        /// <summary>
        /// Fix uses information from GPS satellites and also a differential GPS (DGPS) station.
        /// </summary>
        DGps
    }

    /// <summary>
    /// Type of GPS fix
    /// </summary>
    public enum GpsFixType : int {
        /// <summary>
        /// Type of fix is unknown.
        /// </summary>
        Unknown = 0,
        /// <summary>
        /// Fix is two-dimensional and only provides latitude and longitude. This occurs when only three GPS satellites are used.
        /// </summary>
        TwoD,
        /// <summary>
        /// Fix is three-dimensional and provides latitude, longitude, and elevation. This occurs when four or more GPS satellites are used.
        /// </summary>
        ThreeD
    }

    /// <summary>
    /// Whether 2-D or 3-D mode is selected automatically or manually.
    /// </summary>
    public enum GpsFixSelectionType : int {
        /// <summary>
        /// Selection mode is unknown.
        /// </summary>
        Unknown = 0,
        /// <summary>
        /// Selection mode is automatic.
        /// </summary>
        Auto,
        /// <summary>
        /// Selection mode is manual.
        /// </summary>
        Manual
    }

}
